@@ -30,10 +30,10 @@
 
 #include "config.h"
 #include "gstpub.h"
-#include "alloc.h"
 #include <locale.h>
 #include <stdio.h>
 #include <limits.h>
+#include <stdlib.h>
 #include <string.h>
 #include <langinfo.h>
 #include <iconv.h>
