@@ -32,6 +32,8 @@
 #include "config.h"
 #endif
 
+#include <stdio.h>
+
 #ifdef HAVE_UNISTD_H
 # include <unistd.h>
 #endif 
@@ -71,11 +73,15 @@ getdtablesize(void)
 #else
 #ifdef NOFILE
   return NOFILE;
+#else
+#ifdef FOPEN_MAX
+  return FOPEN_MAX;
 #else
   /* If something went wrong, choose a safe >>>hack<<< value */
   return sizeof(short) * 8;
-#endif
-#endif    
+#endif /* FOPEN_MAX */
+#endif /* NOFILE */
+#endif /* OPEN_MAX */
 }
 #endif
 
