@@ -437,7 +437,7 @@ get_argv (int n)
 PTR
 _gst_dld_open (const char *filename)
 {
-#ifdef HAVE_DLD
+#ifdef ENABLE_DLD
   lt_dlhandle handle;
   void (*initModule) ();
 
