@@ -180,7 +180,7 @@ static const char *feature_strings[] = {
 #ifdef MACHINE_DEFINES
   MACHINE_DEFINES
 #endif
-#ifdef HAVE_DLD
+#ifdef ENABLE_DLD
     "DLD",
 #endif
   NULL
