@@ -264,12 +264,17 @@ extern OOP _gst_nil_oop, _gst_true_oop, _gst_false_oop;
 #ifndef OPTIMIZE
 #define PUSH_OOP(oop) \
   do { \
-    if (!IS_OOP_COPIED (oop)) abort (); \
-    UNCHECKED_PUSH_OOP (oop); \
+    OOP __pushOOP = (oop); \
+    if (IS_OOP (__pushOOP) && !IS_OOP_VALID (__pushOOP)) \
+      abort (); \
+    UNCHECKED_PUSH_OOP (__pushOOP); \
   } while (0)
 #else
 #define PUSH_OOP(oop) \
-    UNCHECKED_PUSH_OOP(oop)
+  do { \
+    OOP __pushOOP = (oop); \
+    UNCHECKED_PUSH_OOP(__pushOOP); \
+  } while (0)
 #endif
 
 #define POP_OOP() \
@@ -287,12 +292,17 @@ extern OOP _gst_nil_oop, _gst_true_oop, _gst_false_oop;
 #ifndef OPTIMIZE
 #define SET_STACKTOP(oop) \
   do { \
-    if (!IS_OOP_COPIED (oop)) abort (); \
-    UNCHECKED_SET_TOP(oop); \
+    OOP __pushOOP = (oop); \
+    if (IS_OOP (__pushOOP) && !IS_OOP_VALID (__pushOOP)) \
+      abort (); \
+    UNCHECKED_SET_TOP(__pushOOP); \
   } while (0)
 #else
 #define SET_STACKTOP(oop) \
-  UNCHECKED_SET_TOP(oop)
+  do { \
+    OOP __pushOOP = (oop); \
+    UNCHECKED_SET_TOP(__pushOOP); \
+  } while (0)
 #endif
 
 #define SET_STACKTOP_INT(i) \
@@ -311,7 +321,7 @@ extern OOP _gst_nil_oop, _gst_true_oop, _gst_false_oop;
   TO_INT(POP_OOP())
 
 #define PUSH_BOOLEAN(exp) \
-  UNCHECKED_PUSH_OOP((exp) ? _gst_true_oop : _gst_false_oop)
+  PUSH_OOP((exp) ? _gst_true_oop : _gst_false_oop)
 
 
 /* Answer whether CLASS is the class that the object pointed to by OOP
