@@ -97,7 +97,7 @@
 
 /* Answer the quantum assigned to each Smalltalk process (in
    milliseconds) before it is preempted.  Setting this to zero
-   disables preemption until gst_processor_scheduler>>#timeSlice: is
+   disables preemption until ProcessorScheduler>>#timeSlice: is
    invoked. */
 #define DEFAULT_PREEMPTION_TIMESLICE 40
 
@@ -1862,7 +1862,7 @@ semaphore_new (int signals)
 }
 
 /* runs before every evaluation (_gst_execute_statements) and before GC turned on.
-   Note that we don't use the incubator because _gst_processor_oop is a global. */
+   We don't use the incubator because _gst_processor_oop is in the root set. */
 void
 _gst_init_process_system (void)
 {
@@ -1885,11 +1885,14 @@ _gst_init_process_system (void)
     processor->processTimeslice =
       FROM_INT (DEFAULT_PREEMPTION_TIMESLICE);
 
+  if (IS_NIL (processor->gcSemaphore))
+    processor->gcSemaphore = semaphore_new (0);
+
   /* No process is active -- so highest_priority_process() need not
      worry about discarding an active process. */
   processor->activeProcess = _gst_nil_oop;
-  activate_process (highest_priority_process ());
   switch_to_process = _gst_nil_oop;
+  activate_process (highest_priority_process ());
   set_preemption_timer ();
 }
 
