@@ -61,32 +61,37 @@
    compile mode */
 mst_Boolean _gst_compile_code = false;
 
-/* The obstack containing parse tree nodes.  */
+/* The obstack containing parse tree nodes. */
 struct obstack *_gst_compilation_obstack;
 
 /* True if errors must be reported to the standard error, false if
    errors should instead stored so that they are passed to Smalltalk
-   code.  */
+   code. */
 mst_Boolean _gst_report_errors = true;
 
 /* The location of the first error reported, stored here so that
-   compilation primitives can pass them to Smalltalk code.  */
+   compilation primitives can pass them to Smalltalk code. */
 char *_gst_first_error_str = NULL;
 char *_gst_first_error_file = NULL;
-int _gst_first_error_line;
+long _gst_first_error_line;
 
 /* The starting-position state.  If this is <0, as soon as another
    token is lexed, the flag is reinitialized to the starting position
-   of the next token.  */
+   of the next token. */
 static int method_start_pos = -1;
 
-/* Answer true if IC is a valid base-10 digit.  */
+/* Answer true if IC is a valid base-10 digit. */
 static mst_Boolean is_digit (int ic);
 
-/* Answer true if C is a valid base-BASE digit.  */
+/* Answer true if C is a valid base-BASE digit. */
 static mst_Boolean is_base_digit (int c,
 				  int base);
 
+/* Answer true if the first thing available in the stream is
+   the string "primitive:" */
+static mst_Boolean parse_primitive (int c,
+				    YYSTYPE * lvalp);
+
 /* Parse the fractional part of a Float constant.  Store it in
    NUMPTR.  Read numbers in base-BASE, the first one being C. 
    Answer the scale (number of digits in numPtr).  If LARGEINTEGER
@@ -104,7 +109,7 @@ static int parse_fraction (int c,
    changed accordingly.  If LARGEINTEGER is not NULL, the
    digits are stored in the buffer maintained by str.c,
    and LARGEINTEGER is set to true if the return value does
-   not have sufficient precision.  */
+   not have sufficient precision. */
 static long double parse_digits (int c,
 			    mst_Boolean negative,
 			    int base,
@@ -113,16 +118,16 @@ static long double parse_digits (int c,
 /* Parse the large integer constant stored as base-BASE
    digits in the buffer maintained by str.c, adjusting
    the sign if NEGATIVE is true.  Return an embryo of the
-   LargeInteger object as a byte_object structure.  */
+   LargeInteger object as a byte_object structure. */
 static byte_object parse_large_integer (mst_Boolean negative,
 					int base);
 
-/* Raise an error.  */
+/* Raise an error. */
 static int invalid (int c,
 		    YYSTYPE * lvalp);
 
 /* Parse a comment.  C is '"'.  Return 0 to indicate the lexer
-   that this lexeme must be ignored.  */
+   that this lexeme must be ignored. */
 static int comment (int c,
 		    YYSTYPE * lvalp);
 
@@ -138,24 +143,24 @@ static int parse_bin_op (int c,
 static int string_literal (int c,
 			   YYSTYPE * lvalp);
 
-/* Parse a number.  C is the first digit.  */
+/* Parse a number.  C is the first digit. */
 static int parse_number (int c,
 			 YYSTYPE * lvalp);
 
-/* Parse an identifier.  C is the first letter.  */
+/* Parse an identifier.  C is the first letter. */
 static int parse_ident (int c,
 			YYSTYPE * lvalp);
 
-/* Try to parse an assignment operator.  C is ':'.  */
+/* Try to parse an assignment operator.  C is ':'. */
 static int parse_colon (int c,
 			YYSTYPE * lvalp);
 
 /* Convert the digit C (if it is a valid base-BASE digit) to its
-   value.  Raise an error if it is invalid.  */
+   value.  Raise an error if it is invalid. */
 static int digit_to_int (int c,
 			 int base);
 
-/* Raise BASE to the N-th power.  */
+/* Raise BASE to the N-th power. */
 static inline long double ipowl (long double base,
 			         int n);
 
@@ -342,7 +347,7 @@ _gst_yylex (PTR lvalp, YYLTYPE *llocp)
          grammar use INTERNAL_TOKEN as the first token of an internal
          method. The same happens when you have a call to methodsFor:
          - INTERNAL_TOKEN is pushed so that the grammar switches to
-         the method_list production.  */
+         the method_list production. */
       _gst_compile_code = false;
       return (INTERNAL_TOKEN);
     }
@@ -487,6 +492,15 @@ parse_bin_op (int c,
   *bp++ = c;
 
   ic = _gst_next_char ();
+  if (c == '<')
+    {
+      if (ic == 'p')
+	{
+	  if (parse_primitive (ic, lvalp))
+	    return (PRIMITIVE_START);
+	}
+    }
+
   if (ic != EOF && (CHAR_TAB (ic)->_gst_char_class & SPECIAL_CHAR))
     *bp++ = ic;			/* accumulate next char */
   else
@@ -500,9 +514,8 @@ parse_bin_op (int c,
 
   lvalp->sval = xstrdup (buf);
 
-  if ((buf[0] == '|' || buf[0] == '<' || buf[0] == '>')
-      && buf[1] == '\0')
-    return (buf[0]);
+  if (strcmp (buf, "|") == 0)
+    return ('|');
 
   else
     return (BINOP);
@@ -541,6 +554,17 @@ string_literal (int c,
   return (STRING_LITERAL);
 }
 
+mst_Boolean
+parse_primitive (int c,
+		 YYSTYPE * lvalp)
+{
+  mst_Boolean result;
+
+  parse_ident (c, lvalp);
+  result = (strcmp (lvalp->sval, "primitive:") == 0);
+  return (result);
+}
+
 int
 parse_ident (int c,
 	     YYSTYPE * lvalp)
@@ -558,7 +582,7 @@ parse_ident (int c,
 	     && (CHAR_TAB (ic)->_gst_char_class & ID_CHAR))
 	_gst_add_str_buf_char (ic);
 
-      /* Read a dot as '::' if followed by a letter.  */
+      /* Read a dot as '::' if followed by a letter. */
       if (ic == '.')
 	{
 	  ic = _gst_next_char ();
@@ -625,19 +649,19 @@ long double
 ipowl (long double base, 
        int n)
 {
-  int k = 1;
-  long double result = 1.0;
-  while (n)
+    int k = 1;
+    long double result = 1.0;
+    while (n)
     {
-      if (n & k)
+        if (n & k)
         {
-          result *= base;
-          n ^= k;
+            result *= base;
+            n ^= k;
         }
-      base *= base;
-      k <<= 1;
+        base *= base;
+        k <<= 1;
     }
-  return result;
+    return result;
 }
 
 int
@@ -700,7 +724,7 @@ parse_number (int c,
 	  /* OOPS...we gobbled the '.' by mistake...it was a statement
 	     boundary delimiter.  We have an integer that we need to
 	     return, and need to push back both the . and the character 
-	     that we just read.  */
+	     that we just read. */
 	  _gst_unread_char (ic);
 	  ic = '.';
 	}
@@ -716,7 +740,7 @@ parse_number (int c,
     do
       {
         /* By default the same as the number of decimal points
-	   we used.  */
+	   we used. */
 	floatExponent = -exponent;
 
 	ic = _gst_next_char ();
@@ -827,12 +851,12 @@ parse_number (int c,
     }
   else if (num >= 0 && num <= 255)
     {
-      lvalp->ival = (intptr_t) num;
+      lvalp->ival = (long) num;
       return (BYTE_LITERAL);
     }
   else
     {
-      lvalp->ival = (intptr_t) num;
+      lvalp->ival = (long) num;
       return (INTEGER_LITERAL);
     }
 }
