@@ -270,7 +270,7 @@ _gst_init_mem (size_t eden, size_t survivor, size_t old,
       if (survivor)
         _gst_tenure_all_survivors ();
 
-      if (old != _gst_mem.old->heap_total)
+      if (old && old != _gst_mem.old->heap_total)
         _gst_grow_memory_to (old);
     }
 
@@ -294,6 +294,9 @@ _gst_init_mem (size_t eden, size_t survivor, size_t old,
   if (big_object_threshold)
     _gst_mem.big_object_threshold = big_object_threshold;
 
+  if (_gst_mem.eden.totalSize < _gst_mem.big_object_threshold)
+    _gst_mem.big_object_threshold = _gst_mem.eden.totalSize;
+
   if (grow_threshold_percent)
     _gst_mem.grow_threshold_percent = grow_threshold_percent;
 
@@ -595,22 +598,32 @@ _gst_swap_objects (OOP oop1,
 		   OOP oop2)
 {
   struct OOP tempOOP;
+
   if (oop2->flags & F_WEAK)
     _gst_make_oop_non_weak (oop2);
 
   if (oop1->flags & F_WEAK)
     _gst_make_oop_non_weak (oop1);
 
-  tempOOP = *oop2;		/* note structure assignment going on here */
-  *oop2 = *oop1;
-  *oop1 = tempOOP;
-
   /* Put the two objects in the same generation.  FIXME: this can be
      a cause of early tenuring, especially since one of them is often
      garbage!  */
   if ((oop1->flags & F_OLD) ^ (oop2->flags & F_OLD))
     _gst_tenure_oop ((oop1->flags & F_OLD) ? oop2 : oop1);
 
+  tempOOP = *oop2;		/* note structure assignment going on here */
+  *oop2 = *oop1;
+  *oop1 = tempOOP;
+
+  /* If the incremental GC has reached oop1 but not oop2 (or vice versa),
+     this flag will end up in the wrong OOP, i.e. in the one that has already
+     been scanned by the incremental GC.  Restore things.  */
+  if ((oop1->flags & F_REACHABLE) ^ (oop2->flags & F_REACHABLE))
+    {
+      oop1->flags ^= F_REACHABLE;
+      oop2->flags ^= F_REACHABLE;
+    }
+
   if (oop2->flags & F_WEAK)
     _gst_make_oop_weak (oop2);
 
@@ -682,15 +695,15 @@ _gst_alloc_obj (size_t size,
      GC, so we use a local var to hold its new value */
   newAllocPtr = _gst_mem.eden.allocPtr + (size >> LONG_SHIFT);
 
+  if UNCOMMON (size >= _gst_mem.big_object_threshold)
+    return _gst_alloc_old_obj (size, p_oop);
+
   if UNCOMMON (newAllocPtr >= _gst_mem.eden.maxPtr)
     {
       _gst_scavenge ();
       newAllocPtr = _gst_mem.eden.allocPtr + size;
     }
 
-  if UNCOMMON (size >= _gst_mem.big_object_threshold)
-    return _gst_alloc_old_obj (size, p_oop);
-
   p_instance = (mst_Object) _gst_mem.eden.allocPtr;
   _gst_mem.eden.allocPtr = newAllocPtr;
   *p_oop = alloc_oop (p_instance, _gst_mem.active_flag);
@@ -749,9 +762,6 @@ mst_Object _gst_alloc_words (size_t size)
       abort ();
     }
 
-  if UNCOMMON (size >= _gst_mem.big_object_threshold)
-    abort ();
-
   p_instance = (mst_Object) _gst_mem.eden.allocPtr;
   _gst_mem.eden.allocPtr = newAllocPtr;
   p_instance->objSize = FROM_INT (size);
@@ -1134,7 +1144,7 @@ _gst_finish_incremental_gc ()
 
   for (oop = _gst_mem.highest_swept_oop, firstOOP = _gst_mem.last_swept_oop;
        --oop > firstOOP; oop->flags &= ~F_REACHABLE)
-    if (!IS_OOP_VALID(oop))
+    if (!IS_OOP_VALID_GC(oop))
       {
         _gst_sweep_oop (oop);
 	_gst_mem.num_free_oops++;
@@ -1180,7 +1190,7 @@ _gst_incremental_gc_step ()
 	  break;
 	}
 
-      if (!IS_OOP_VALID(oop))
+      if (!IS_OOP_VALID_GC(oop))
         {
           i++;
           _gst_sweep_oop (oop);
@@ -1199,7 +1209,7 @@ reset_incremental_gc (OOP firstOOP)
 
   /* This loop is the same as that in alloc_oop.  Skip low OOPs
      that are allocated */
-  for (oop = firstOOP; IS_OOP_VALID (oop); oop->flags &= ~F_REACHABLE, oop++)
+  for (oop = firstOOP; IS_OOP_VALID_GC (oop); oop->flags &= ~F_REACHABLE, oop++)
 #if defined(USE_JIT_TRANSLATION)
     if (oop->flags & F_XLAT)
       {
@@ -1228,7 +1238,7 @@ reset_incremental_gc (OOP firstOOP)
   _gst_finish_incremental_gc ();
 #else
   /* Skip high OOPs that are unallocated.  */
-  for (oop = _gst_mem.last_allocated_oop; !IS_OOP_VALID (oop); oop--)
+  for (oop = _gst_mem.last_allocated_oop; !IS_OOP_VALID_GC (oop); oop--)
     _gst_sweep_oop (oop);
 
   _gst_mem.last_allocated_oop = oop;
@@ -1298,17 +1308,27 @@ _gst_sweep_oop (OOP oop)
 void
 mourn_objects (void)
 {
-  OOP *pOOP;
+  mst_Object array;
   long size;
+  gst_processor_scheduler processor;
 
-  /* Make a local copy of the buffer */
-  size = _gst_buffer_size ();
-  pOOP = alloca (size);
-  _gst_copy_buffer (pOOP);
-  size /= SIZEOF_CHAR_P;
+  size = _gst_buffer_size () / sizeof (OOP);
+  if (!size)
+    return;
 
-  while (size--)
-    _gst_msg_send (*pOOP++, _gst_mourn_symbol, NULL);
+  processor = (gst_processor_scheduler) OOP_TO_OBJ (_gst_processor_oop);
+  if (!IS_NIL (processor->gcArray))
+    {
+      _gst_errorf ("Too many garbage collections, finalizers missed!");
+      _gst_errorf ("This is a bug, please report.");
+    }
+  else
+    {
+      /* Copy the buffer */
+      array = new_instance_with (_gst_array_class, size, &processor->gcArray);
+      _gst_copy_buffer (array->data);
+      _gst_async_signal (processor->gcSemaphore);
+    }
 }
 
 
@@ -1501,7 +1521,7 @@ check_weak_refs ()
           if (IS_INT (oop))
 	    continue;
 
-          if (!IS_OOP_VALID (oop))
+          if (!IS_OOP_VALID_GC (oop))
             {
               mourn = true;
 	      *field = _gst_nil_oop;
@@ -1603,7 +1623,7 @@ scan_grey_pages ()
 	  printf ("Protecting %p\n", node->base);
 #endif
 	  _gst_mem.rememberedTableEntries--;
-	  _gst_mem_protect ((PTR) node->base, getpagesize(), PROT_READ);
+	  _gst_mem_protect ((PTR) node->base, node->n * sizeof(OOP), PROT_READ);
 	  *next = node->next;
 	  xfree (node);
 	}
