@@ -29,14 +29,8 @@
  ***********************************************************************/
 
 
-#include "gst.h"
 #include "gstpriv.h"
 
-#if STDC_HEADERS
-#include <string.h>		/* for memcpy */
-#include <stdlib.h>
-#endif /* STDC_HEADERS */
-#include <stdio.h>
 
 /* Define this to disable the peephole bytecode optimizer.  It works
    well and increases a bit performance, so there's no reason to do that
@@ -166,7 +160,7 @@ static mst_Boolean optimize_basic_block (gst_uchar * from,
 
 /* This compares two block_boundary structures according to their
    bytecode position. */
-static int compare_blocks (const PTR a, const PTR b) FN_PURE;
+static int compare_blocks (const PTR a, const PTR b) ATTRIBUTE_PURE;
 
 /* This answers how the dirtyness of BLOCKCLOSUREOOP affects
    the block that encloses it.  */
