@@ -104,6 +104,8 @@ typedef struct printf_info
   unsigned showsign:1;
   unsigned group:1;
   unsigned wide:1;
+
+  const union printf_arg *args;
 } printf_info;
 
 /**
@@ -164,7 +166,6 @@ typedef int printf_arginfo_function (struct printf_info *pparser, size_t n, int
 /**
  * spec_entry:
  * @spec: the specifier character that was matched
- * @modifier_char: usually %FALSE, these specifiers aren't influenced by $ and cannot be overridden
  * @type: when @arg is NULL, the type of the only argument to the specifier
  * @fmt: the handler function to actually print the arguments to the specifier
  * @arg: the handler function to tell %printf about the types of the arguments to the specifier
@@ -175,7 +176,7 @@ typedef int printf_arginfo_function (struct printf_info *pparser, size_t n, int
 typedef struct spec_entry
 {
   int spec;
-  int modifier_char;
+  int unused;  /* for binary compatibility */
   int type;
   printf_function *fmt;
   printf_arginfo_function *arg;
@@ -265,6 +266,7 @@ enum
       } \
   } SNV_STMT_END
 
+#line 269 "./printf.in"
 /**
  * printf_generic_info:   
  * @pinfo: the current state information for the format
@@ -300,7 +302,7 @@ extern int printf_generic_info (struct printf_info *const pinfo, size_t n, int *
 extern int printf_generic (STREAM *stream, struct printf_info *const pinfo, union printf_arg const *args);
 
 
-#line 268 "./printf.in"
+#line 270 "./printf.in"
 /**
  * register_printf_function:  
  * @spec: the character which will trigger @func, cast to an unsigned int.
@@ -787,7 +789,7 @@ extern int snv_vasprintf (char **result, const char *format, va_list ap);
 extern int snv_asprintfv (char **result, const char *format, snv_constpointer const args[]);
 
 
-#line 269 "./printf.in"
+#line 271 "./printf.in"
 
 /* If you don't want to use snprintfv functions for *all* of your string
    formatting API, then define COMPILING_SNPRINTFV_C and use the snv_
